import unittest
from naclo import neutralize
from naclo import mol_conversion


class TestNeutralize(unittest.TestCase):
    def test_init_neutralization_rxns(self):
        rxns = neutralize.init_neutralization_rxns()
        rxns = list(rxns.values()) +  list(rxns.keys())
        
        self.assertEqual(
            mol_conversion.mols_2_smiles(rxns),
            ['O', '*']
        )  # What does test_init_neutralization do????
        

    def test_neutralize_charges(self):
        # Doesn't do anything yet
        pass


if __name__ == '__main__':
    unittest.main()



import unittest
from naclo import mol_conversion
from rdkit.Chem import PandasTools
from rdkit import DataStructs
import pandas as pd
import numpy as np


class TestMolConversion(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Load excel test data
        test_excel = pd.read_excel('test/assets/excel_test_case.xlsx')
        cls.excel_smiles = list(test_excel.Smiles)
        cls.excel_inchi_keys = list(test_excel.InChi)

        # Load SDF test data
        test_sdf = PandasTools.LoadSDF('test/assets/sdf_test_case.sdf', molColName='Molecule')
        cls.sdf_mols = list(test_sdf.Molecule)
        cls.sdf_smiles = list(test_sdf.Smiles)
        
        return super().setUpClass()
    
    def test_mols_2_smiles(self):
        self.assertEqual(
            self.sdf_smiles, 
            mol_conversion.mols_2_smiles(self.sdf_mols)
        )
        
    def test_smiles_2_mols(self):
        self.assertEqual(
            self.excel_smiles,
            mol_conversion.mols_2_smiles(mol_conversion.smiles_2_mols(self.excel_smiles))
        )
        
    def test_mols_2_inchi_keys(self):
        self.assertEqual(
            self.excel_inchi_keys,
            mol_conversion.mols_2_inchi_keys(mol_conversion.smiles_2_mols(self.excel_smiles))
        )
        
    def test_smiles_2_inchi_keys(self):
        self.assertEqual(
            self.excel_inchi_keys, 
            mol_conversion.smiles_2_inchi_keys(self.excel_smiles)
        )

    def test_mols_2_ecfp(self):
        ecfp = mol_conversion.mols_2_ecfp(self.sdf_mols, return_numpy=True)
        self.assertEqual(
            len(ecfp[0]),
            1024
        )
        self.assertIsInstance(
            ecfp[0],
            np.ndarray
        )
        
        print_objs = mol_conversion.mols_2_ecfp(self.sdf_mols, return_numpy=False)
        self.assertIsInstance(
            print_objs[0],
            DataStructs.cDataStructs.UIntSparseIntVect
        )
        
    def test_mols_2_maccs(self):
        maccss = mol_conversion.mols_2_maccs(self.sdf_mols)
        self.assertIsInstance(
            maccss[0],
            DataStructs.cDataStructs.ExplicitBitVect
        )
        
    def test_mols_2_ecfp_plus_descriptors(self):
        other_df = pd.DataFrame({
            'other1': [1, 3, 9, 22],
            'other2': [0.4, 0.77, 0.9, 0.01]
        })
        
        # Z norm
        out = mol_conversion.mols_2_ecfp_plus_descriptors(self.sdf_mols, other_df)


if __name__ == '__main__':
    unittest.main()



import pandas as pd
import warnings
from typing import Callable, Dict, Union, Optional

# sourced from github.com/jwgerlach00
import naclo
import stse
from naclo.__asset_loader import recognized_bleach_options as recognized_options
from naclo.__asset_loader import bleach_default_params as default_params
from naclo.__asset_loader import bleach_default_options as default_options
from naclo.__naclo_util import recognized_options_checker


class Bleach:
    filter_fragments_methods = ['carbon_count', 'mw', 'atom_count', 'none']
    
    def __init__(self, df:pd.DataFrame, params:dict=default_params, options:dict=default_options) -> None:  # *
        # Load user options
        self.mol_settings = options['molecule_settings']
        self.file_settings = options['file_settings']
        recognized_options_checker(options, recognized_options)

        self.__recognized_structures = ['smiles', 'mol']
        self.__default_cols = {
            'smiles': 'SMILES',
            'mol': 'ROMol',
            'inchi_key': 'InchiKey',
            'mw': 'MW'
        }

        # Save user input data
        self.original_df = df.copy()
        self.df = df.copy()

        # Load file parameters
        self.structure_col = params['structure_col']
        self.structure_type = params['structure_type']
        self.target_col = params['target_col']
        self.__param_checker()

        self.mol_col = None
        self.smiles_col = None
        self.inchi_key_col = None
        self.__set_structure_cols()  # Assign mol and SMILES cols using input + defaults
        
        # Set staticmethods to instance methods
        self.convert_units = self.__instance_convert_units
        self.mol_cleanup = self.__instance_mol_cleanup
        self.handle_duplicates = self.__instance_handle_duplicates
        self.append_columns = self.__instance_append_columns
        self.remove_header_chars = self.__instance_remove_header_chars


# -------------------------------------------------- ERROR CHECKING -------------------------------------------------- #
    def __param_checker(self) -> None:  # *
        """Checks for errors related to declared parameters.

        Raises:
            ValueError: NO_STRUCTURE_COLUMN
            ValueError: STRUCTURE_COLUMN_NOT_FOUND
            ValueError: INVALID_STRUCTURE_TYPE
            ValueError: TARGET_COLUMN_NOT_FOUND
        """
        if not self.structure_col:
            raise ValueError('NO_STRUCTURE_COLUMN', 'Must specify the name of the structure column in params')

        if self.structure_col not in self.df.columns:
            raise ValueError('STRUCTURE_COLUMN_NOT_FOUND', f'The structure column: "{self.structure_col}"" is not \
                present in the data: "{list(self.df.columns)}"')

        if self.structure_type not in self.__recognized_structures:
            raise ValueError('INVALID_STRUCTURE_TYPE', f'Structure type: "{self.structure_type}"" is not one of: \
                {self.__recognized_structures}')

        if self.target_col and self.target_col not in self.df.columns:
            raise ValueError('TARGET_COLUMN_NOT_FOUND', f'The target column: "{self.target_col}"" is not present in \
                the data: "{list(self.df.columns)}"')


# -------------------------------------------------- PRIVATE METHODS ------------------------------------------------- #
    def __set_structure_cols(self) -> None:
        """Sets Mol and SMILES columns using declared structure type."""
        self.mol_col = self.structure_col if self.structure_type == 'mol' else self.__default_cols['mol']
        self.smiles_col = self.structure_col if self.structure_type == 'smiles' else self.__default_cols['smiles']
        self.inchi_key_col = self.__default_cols['inchi_key']

    def __drop_na_structures(self) -> None:
        """Drops NA along declared structure column."""
        self.df.dropna(subset=[self.structure_col], inplace=True)
        if not len(self.df):
            warnings.warn('ALL_NA_STRUCTURES: All structures in specified column were NA, all rows dropped',
                          RuntimeWarning)

    def __drop_na_targets(self) -> None:
        """Drops NA along declared target column"""
        run_na_targets = self.file_settings['remove_na_targets']['run']

        if self.target_col and run_na_targets and len(self.df):  # If run and TARGET COLUMN DECLARED
            self.df.dropna(subset=[self.target_col], inplace=True)
            if not len(self.df):
                warnings.warn('ALL_NA_TARGETS: All targets in specified column were NA, all rows dropped',
                              RuntimeWarning)

        elif run_na_targets:  # If run but not declared target
            warnings.warn('NA_TARGETS: options.file_settings.remove_na_targets was set to run but no activity column \
                was specified', RuntimeWarning)

    def __build_smiles(self) -> None:
        """Creates a SMILES column in the dataset using dataset MolFile column. DROPS NA."""
        self.df = naclo.dataframes.df_mols_2_smiles(self.df, self.mol_col, self.smiles_col)

    def __build_mols(self) -> None:
        """Creates MolFile column in the dataset using dataset SMILES column. DROPS NA."""
        self.df = naclo.dataframes.df_smiles_2_mols(self.df, self.smiles_col, self.mol_col)
        
    @staticmethod
    def __filter_fragments_factory(filter:str) -> Callable:
        """Returns a callable SMILES fragment filter function using a key.

        Args:
            filter (str): Function key.

        Raises:
            ValueError

        Returns:
            Callable: Filter function.
        """
        if filter == 'carbon_count':
            return naclo.fragments.carbon_count
        elif filter == 'mw':
            return naclo.fragments.mw
        elif filter == 'atom_count':
            return naclo.fragments.atom_count
        else:
            raise ValueError('Filter method is not allowed')

    @staticmethod
    def __remove_fragments(df:pd.DataFrame, smiles_col_name:str, mol_col_name:str, salts:bool,
                           filter_method:Optional[str]) -> pd.DataFrame:
        """Removes salts if specified. Drops NA as a result of salt removal. Filters out other fragments by specified
        method."""
        df = df.copy()
        rebuild_mols = lambda x: naclo.dataframes.df_smiles_2_mols(x, smiles_col_name, mol_col_name)
            
        if salts:
            df[smiles_col_name] = df[smiles_col_name].apply(naclo.fragments.remove_recognized_salts)
            df = rebuild_mols(df)

            # Drop NA (blank string after salts)
            df = stse.dataframes.convert_to_nan(df, na=[''])  # Convert bc NA is just empty string
            df.dropna(subset=[smiles_col_name], inplace=True)  # Drop NA bc may include molecule that is ONLY salts

        # Filter
        if filter_method and filter_method != 'none':
            df[smiles_col_name] = df[smiles_col_name].apply(Bleach.__filter_fragments_factory(filter_method))
            df = rebuild_mols(df)
            
        return df

    @staticmethod
    def __append_inchi_keys(df:pd.DataFrame, mol_col_name:str, inchi_key_col_name:str) -> pd.DataFrame:
        """Declares inchi key column name using default. Appends inchi keys to dataset."""
        return naclo.dataframes.df_mols_2_inchi_keys(df, mol_col_name, inchi_key_col_name)

    @staticmethod
    def __drop_columns(df:pd.DataFrame, column_mapper:Dict[str, bool], mol_col_name:str, smiles_col_name:str,
                       inchi_key_col_name:str) -> pd.DataFrame:
        """Removes columns that the user does not want in the final output."""
        
        # Drop added columns from built if not requested
        if not column_mapper['mol']:
            df = df.drop(mol_col_name, axis=1)
        if not column_mapper['inchi_key']:
            df = df.drop(inchi_key_col_name, axis=1)
        if not column_mapper['smiles']:
            df = df.drop(smiles_col_name, axis=1)
        
        return df

    @staticmethod
    def __add_columns(df:pd.DataFrame, column_mapper:Dict[str, bool], mol_col_name:str) -> pd.DataFrame:
        """Add columns that the user wants in the final output."""
        # Add MW column
        if column_mapper['mw']:
            df = df.assign(MW = naclo.mol_weights(df[mol_col_name]))
        return df


# ---------------------------------------------------- MAIN STEPS ---------------------------------------------------- #
    # Step 1
    def drop_na(self) -> None:  # *
        """Converts blanks to NA. Drops NA Mols or SMILES. Handles NA targets. Removes entire NA columns"""

        # Convert all df blanks and 'none' to NA
        self.df = stse.dataframes.convert_to_nan(self.df)

        # Drop rows
        self.__drop_na_structures()
        self.__drop_na_targets()

        # Drop cols
        self.df = stse.dataframes.remove_nan_cols(self.df)  # After dropping rows because columns may BECOME empty

    # Step 2 
    def init_structure_compute(self) -> None:  # *
        """Builds (or rebuilds from Mols) SMILES. Builds Mols if not present in dataset."""
        if self.structure_type == 'mol':
            self.__build_smiles()
            # Rebuilding Mols not necessary

        elif self.structure_type == 'smiles':
            self.__build_mols()
            self.__build_smiles()  # Canonicalize SMILES
    
    # Step 3
    @staticmethod
    def convert_units(df:pd.DataFrame, mol_col_name:str, value_col_name:str, units_col_name:str,
                      output_units:str, drop_na_units:bool) -> pd.DataFrame:
        df = df.copy()
        
        uc = naclo.UnitConverter(df[value_col_name], df[units_col_name], naclo.mol_weights(df[mol_col_name]))
        
        if output_units == 'molar':
            col_name = f'molar_{units_col_name}'
            df[col_name] = uc.to_molar()
        elif output_units == 'neg_log_molar':
            col_name = f'neg_log_molar_{units_col_name}'
            df[col_name] = uc.to_neg_log_molar()
        else:
            raise ValueError(f'Output units: "{output_units}" are not recognized')
        
        return df.dropna(subset=[col_name]) if drop_na_units else df
    
    def __instance_convert_units(self):
        convert_units = self.mol_settings['convert_units']
        if convert_units['units_col']:
            if not self.target_col:
                warnings.warn('CONVERT_UNITS: options.molecule_settings.convert_units was set to run but no activity \
                    column was specified', RuntimeWarning)
                return
            self.df = Bleach.convert_units(df=self.df, mol_col_name=self.mol_col, value_col_name=self.target_col,
                                           units_col_name=convert_units['units_col'],
                                           output_units=convert_units['output_units'],
                                           drop_na_units=convert_units['drop_na'])

    # Step 4
    @staticmethod
    def mol_cleanup(df:pd.DataFrame, smiles_col_name:str, mol_col_name:str, run_salts:bool, filter_method:Optional[str],
                    run_neutralize:bool) -> pd.DataFrame:  # * (except neutralize)
        """Cleans Mols and SMILES."""
        df = df.copy()

        # Step 1: Deal with fragments (includes salt step -- may include a molecule that is ONLY salts (NA dropped))
        df = Bleach.__remove_fragments(df, smiles_col_name, mol_col_name, run_salts,
                                       filter_method=(None if filter_method == 'none' else filter_method))

        # Step 2: Neutralize mols
        if run_neutralize:
            df[mol_col_name] = naclo.neutralize.neutralize_charges(df[mol_col_name])
            df = naclo.dataframes.df_mols_2_smiles(df, mol_col_name, smiles_col_name)  # Rebuild SMILES
        
        return df
    
    def __instance_mol_cleanup(self) -> None:
        self.df = Bleach.mol_cleanup(self.df, self.smiles_col, self.mol_col,
                                     run_salts=self.mol_settings['remove_fragments']['salts'],
                                     filter_method=self.mol_settings['remove_fragments']['filter_method'],
                                     run_neutralize=self.mol_settings['neutralize_charges']['run'])

    # Step 5
    @staticmethod
    def handle_duplicates(df:pd.DataFrame, mol_col_name:str, inchi_key_col_name:str, target_col:Union[str, None]=None,
                          method='average') -> pd.DataFrame:  # *
        """Computes inchi keys. Averages, removes, or keeps duplicates. ONLY BY INCHI KEY FOR NOW."""
        df = Bleach.__append_inchi_keys(df, mol_col_name, inchi_key_col_name)

        if method == 'average' and target_col:
            df = stse.duplicates.average(df, subsets=[inchi_key_col_name], average_by=target_col)
        elif method == 'remove' or (method == 'average' and not target_col):
            df = stse.duplicates.remove(df, subsets=[inchi_key_col_name])
        return df
    
    def __instance_handle_duplicates(self) -> None:
        self.df = Bleach.handle_duplicates(self.df, self.mol_col, self.inchi_key_col, self.target_col,
                                           method=self.file_settings['duplicate_compounds']['selected'])

    # Step 6
    @staticmethod
    def append_columns(df:pd.DataFrame, column_mapper:Dict[str, bool], mol_col_name:str, smiles_col_name:str,
                       inchi_key_col_name:str) -> pd.DataFrame:  # *
        """Drops and adds columns depending on what the user wants returned.

        Args:
            df (pandas DataFrame): Data to transform
        """
        df = df.copy()
        
        df = Bleach.__drop_columns(df, column_mapper, mol_col_name, smiles_col_name, inchi_key_col_name)
        df = Bleach.__add_columns(df, column_mapper, mol_col_name)
        return df
    
    def __instance_append_columns(self) -> None:
        self.df = Bleach.append_columns(self.df, self.file_settings['append_columns'], self.mol_col, self.smiles_col,
                                        self.inchi_key_col)

    # Step 7
    @staticmethod
    def remove_header_chars(df, chars) -> pd.DataFrame:  # *
        """Removes any chars listed in a string of chars from the df column headers.
        """
        return stse.dataframes.remove_header_chars(df, chars)
    
    def __instance_remove_header_chars(self) -> None:
        self.df = Bleach.remove_header_chars(self.df, self.file_settings['remove_header_chars']['chars'])


# ----------------------------------------------------- MAIN LOOP ---------------------------------------------------- #
    def main(self) -> pd.DataFrame:
        """Main bleach loop.

        Returns:
            pandas DataFrame: Cleaned df
        """
        self.drop_na()  # Before init_structure bc need NA
        self.init_structure_compute()
        self.convert_units()
        self.mol_cleanup()
        self.handle_duplicates()
        self.append_columns()
        self.remove_header_chars()
        return self.df



from rdkit import Chem
from warnings import warn
from rdkit.Chem import PandasTools
import pandas as pd
from typing import Any, Callable, Optional, Union, IO
import numpy as np

from naclo.Writer import Writer  # Nested import (not present in __init__)


def __exception_2_nan(x:Any, func:Callable) -> float:
    try:
        return func(x)
    except Exception:
        return np.nan
    
def df_mols_2_smiles(df:pd.DataFrame, mol_name:str, smiles_name:str, dropna:bool=True) -> pd.DataFrame:  # *
    """Adds SMILES Key column to df using Mol column as reference.

    Args:
        df (pandas DataFrame): DataFrame to add SMILES column to.
        mol_name (str): Name of Mol column in df.
        inchi_name (str): Name of InChi column in df.
        dropna (bool, optional): Drop NA SMILES. Defaults to True.

    Returns:
        pandas DataFrame: DataFrame with SMILES column appended.
    """
    df = df.copy()
    df[smiles_name] = df[mol_name].map(lambda x: __exception_2_nan(x, Chem.MolToSmiles), na_action='ignore')
    return df.dropna(subset=[smiles_name]) if dropna else df

def df_smiles_2_mols(df:pd.DataFrame, smiles_name:str, mol_name:str, dropna:bool=True) -> pd.DataFrame:  # *
    """Adds rdkit Mol column to df using SMILES column as reference.

    Args:
        df (pandas DataFrame): DataFrame to add Mol column to.
        smiles_name (str): Name of SMILES column in df.
        molecule_name (str): Name of Mol column in df.
        dropna (bool, optional): Drop NA Mols. Defaults to True.

    Returns:
        pandas DataFrame: DataFrame with Mol column appended.
    """
    df = df.copy()
    df[mol_name] = df[smiles_name].map(lambda x: __exception_2_nan(x, Chem.MolFromSmiles), na_action='ignore')
    return df.dropna(subset=[mol_name]) if dropna else df

def df_mols_2_inchi_keys(df:pd.DataFrame, mol_name:str, inchi_name:str, dropna:bool=True) -> pd.DataFrame:  # *
    """Adds InChi Key column to df using Mol column as reference.
    
    Args:
        df (pandas DataFrame): DataFrame to add InChi column to.
        mol_name (str): Name of InChi column in df.
        inchi_name (str): Name of InChi column in df.
        dropna (bool, optional): Drop NA InChis. Defaults to True.

    Returns:
        pandas DataFrame: DataFrame with InChi column appended.
    """
    df = df.copy()
    df[inchi_name] = df[mol_name].map(lambda x: __exception_2_nan(x, Chem.MolToInchiKey), na_action='ignore')
    return df.dropna(subset=[inchi_name]) if dropna else df

def df_smiles_2_inchi_keys(df:pd.DataFrame, smiles_name:str, inchi_name:str, dropna:bool=True) -> pd.DataFrame:
    df = df.copy()
    df = df_smiles_2_mols(df, smiles_name, 'ROMol', dropna=dropna)
    df = df_mols_2_inchi_keys(df, 'ROMol', inchi_name, dropna=dropna)
    return df.drop(columns=['ROMol'])  # Remove added Mol column

def write_sdf(df, out_path:Union[str, IO], mol_col_name:str, id_column_name:str='RowID') -> None:  # *
    """Writes dataframe to SDF file. Includes ID name if ID is valid.

    Args:
        df (DataFrame): Data to write.
        out_path (str or file-like): Path to save SDF to.
        mol_col_name (str): Name of molecule column in dataframe.
        id_column_name (str, optional): Name of id column. Defaults to 'ID'.
    """
    try:
        PandasTools.WriteSDF(df, out_path, molColName=mol_col_name, properties=df.columns, idName=id_column_name)
    except KeyError:
        PandasTools.WriteSDF(df, out_path, molColName=mol_col_name, properties=df.columns, idName='RowID')
        warn(f'write_sdf \'{id_column_name}\' ID name invalid', UserWarning)

def id_mol_col(df:pd.DataFrame) -> Optional[pd.Index]:
    """Identifies first column that contains a Mol object based on first row.

    Args:
        df (pd.DataFrame): Input data.

    Returns:
        Union[None, pd.Index]: Index of first Mol column if found. Else None.
    """
    for i in range(len(df)):
        df.iloc[i]
        first_row = [isinstance(m, Chem.rdchem.Mol) for m in df.iloc[i]]
        if sum(first_row) > 0:
            return df.columns[first_row.index(True)]
        
    return None



import unittest
import pandas as pd
import numpy as np
from copy import deepcopy
from rdkit.Chem import MolFromSmiles
from rdkit.Chem.Descriptors import ExactMolWt
from math import log10

from naclo import binarize_default_params, binarize_default_options
from naclo import Binarize


class TestBinarize(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.default_params = binarize_default_params
        cls.default_options = binarize_default_options
        
        cls.test_df = pd.DataFrame({
            'smiles': [
                'CCC',
                'C',
                'CN=C=O',
                'CN(C)C.Cl',
                None
            ],
            'target': [
                55,
                4,
                7,
                100,
                2000
            ],
            'units': [
                'ug•ml-1',
                'mg/l',
                'unrecognized',
                np.nan,
                'pm'
            ],
            'qualifiers': [
                '>',
                '<',
                '=',
                '',
                '<'
            ]
        })
        
        cls.default_params['structure_col'] = 'smiles'
        cls.default_params['structure_type'] = 'smiles'
        cls.default_params['target_col'] = 'target'
        cls.default_params['decision_boundary'] =  7  # neg log molar
        
        return super().setUpClass()
    
    def test_convert_units(self):
        options = deepcopy(self.default_options)
        options['convert_units']['units_col'] = 'units'
        
        first_two_rows = self.test_df.iloc[:2]
        
        mws = [ExactMolWt(MolFromSmiles(smi)) for smi in first_two_rows['smiles']]
        targets = first_two_rows['target']
        conversion_factor = 1e-3  # ug/ml and mg/l are the same
        
        first_two_expected = (targets*conversion_factor / mws).tolist()
        
        expected_molar = first_two_expected + 2*[np.nan]
        
        binarize = Binarize(self.test_df, params=self.default_params, options=options)
        
        # Molar
        self.assertTrue(
            np.allclose(
                binarize.convert_units('molar'),
                expected_molar, 
                equal_nan=True
            )
        )
        
        # -log(Molar)
        self.assertTrue(
            np.allclose(
                binarize.convert_units('neg_log_molar'),
                [-1*log10(x) for x in expected_molar],
                equal_nan=True
            )
        )
        
        # Nanomolar
        self.assertTrue(
            np.allclose(
                binarize.convert_units('nanomolar'),
                [1e9*x for x in expected_molar],
                equal_nan=True
            )
        )
        
        # Unknown units
        with self.assertRaises(ValueError):
            binarize.convert_units('unknown')
        
    def test_binarize_no_qualifiers(self):
        options = deepcopy(self.default_options)
        options['qualifiers']['run'] = False
        
        expected_arrs = {
            '>': [1, 0, 0, 1, 1],
            '<': [0, 1, 0, 0, 0],
            '>=': [1, 0, 1, 1, 1],
            '<=': [0, 1, 1, 0, 0]
        }
        
        # Loop through allowed operators
        for options['active_operator'] in ['>', '<', '>=', '<=']:
            binarize = Binarize(self.test_df, params=self.default_params, options=options)
            out_df, out_arr = binarize.binarize(self.test_df['target'])

            self.assertTrue(
                out_df.equals(binarize.df)  # No qualifier drop bc no qualifiers
            )
            
            self.assertTrue(
                np.array_equal(
                    list(expected_arrs[options['active_operator']]),
                    out_arr
                )
            )

        # Unknown operator
        options['active_operator'] = 'unknown'
        with self.assertRaises(ValueError):
            binarize = Binarize(self.test_df, params=self.default_params, options=options)
            binarize.binarize(self.test_df['target'])
        
    def test_binarize_with_qualifiers(self):
        options = deepcopy(self.default_options)
        options['qualifiers']['run'] = True
        options['qualifiers']['qualifier_col'] = 'qualifiers'
        
        expected_arrs = {
            '>': [1, 0, 0],
            '<': [0, 1, 0],
            '>=': [1, 0, 1],
            '<=': [0, 1, 1]
        }
        
        # Loop through allowed operators
        for options['active_operator'] in ['>', '<', '>=', '<=']:
            binarize = Binarize(self.test_df, params=self.default_params, options=options)
            out_df, out_arr = (binarize.binarize(self.test_df['target'][:-1]))
            
            self.assertTrue(
                out_df.equals(binarize.df.dropna(subset=['qualifiers']))
            )
            
            self.assertTrue(
                np.array_equal(
                    list(expected_arrs[options['active_operator']]),
                    out_arr
                )
            )
        
        # Unknown operator
        options['active_operator'] = 'unknown'
        with self.assertRaises(ValueError):
            binarize = Binarize(self.test_df, params=self.default_params, options=options)
            binarize.binarize(self.test_df['target'][:-1])
            
    def test_handle_duplicates(self):
        input_df = pd.DataFrame({
            'smiles': ['CC', 'CC', 'CCC', 'CCC', 'CCC', 'CCC', 'CCC', 'O=S=O', 'O=S=O', 'O=S=O', 'O'],
            'bin': [0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 1]
        })
        
        # Agree ratio of 0.8
        expected_df = pd.DataFrame({
            'smiles': ['CC', 'CCC', 'O'],  # SO2 dropped bc bad ratio
            'bin': [0, 1, 1]
        })
        out = Binarize.handle_duplicates(input_df, 'smiles', 'smiles', 'bin', agree_ratio=0.8)
        self.assertTrue(
            out.equals(expected_df)
        )
        
        # Agree ratio of 0.3
        expected_df = pd.DataFrame({
            'smiles': ['CC', 'CCC', 'O=S=O', 'O'],
            'bin': [0, 1, 0, 1]
        })
        out = Binarize.handle_duplicates(input_df, 'smiles', 'smiles', 'bin', agree_ratio=0.6)
        self.assertTrue(
            out.equals(expected_df)
        )
        
    def test_main(self):
        options = {
            'duplicates': {
                'run': False,
                'agree_ratio': 0.8
            },
            'convert_units': {
                'units_col': 'units',
                'output_units': 'molar'
            },
            'qualifiers': {
                'run': True,
                'qualifier_col': 'qualifiers'
            },
            'active_operator': '<=',
            'drop_na': False
        }

        # Drop NA = False
        binarize = Binarize(self.test_df, params=self.default_params, options=options)
        out = binarize.main()
        
        self.assertEqual(
            out['smiles'].tolist(),
            ['CCC', 'C', 'CN=C=O']
        )
        self.assertEqual(
            out['target'].tolist(),
            [55, 4, 7]
        )
        self.assertEqual(
            out['units'].tolist(),
            ['ug•ml-1', 'mg/l', 'unrecognized']
        )
        self.assertEqual(
            out['qualifiers'].tolist(),
            ['>', '<', '=']
        )
        self.assertTrue(
            np.allclose(
                out['molar_target'].tolist(),
                [0.001248224110253472, 0.0002495118903683718, np.nan],
                equal_nan=True
            )
        )
        self.assertTrue(
            np.allclose(
                out['binarized_target'].tolist(),
                [np.nan, 1, np.nan],
                equal_nan=True
            )
        )
        
        # Drop NA = True
        options['drop_na'] = True
        binarize = Binarize(self.test_df, params=self.default_params, options=options)
        out = binarize.main()
        
        expected = pd.DataFrame({
            'smiles': ['C'],
            'target': [4],
            'units': ['mg/l'],
            'qualifiers': ['<'],
            'molar_target': [0.0002495118903683718],
            'binarized_target': [1.0]
        })
        
        self.assertTrue(
            out.reset_index(drop=True).equals(expected)
        )


if __name__ == '__main__':
    unittest.main()



from ast import Bytes
import unittest
import pandas as pd
from rdkit import Chem
from rdkit.Chem import PandasTools
from setuptools import setup
from io import BytesIO

import naclo
from naclo import dataframes


class TestDataframes(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_df = pd.DataFrame({'a': [4, '5', '6'], 'b': ['6', '5', '4']})
        cls.mol_test_df = pd.DataFrame({
            'SMILES': ['O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1', 
                       'O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1', 
                       'CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O',
                       None],
            'Molecule': [Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1'),
                         Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1'),
                         Chem.MolFromSmiles('CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O'),
                         None],
            'InChi': [Chem.MolToInchiKey(Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1')),
                      Chem.MolToInchiKey(Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1')),
                      Chem.MolToInchiKey(Chem.MolFromSmiles('CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O')),
                      None]
            })
        
        cls.excel_df = pd.read_excel('test/assets/excel_test_case.xlsx')
        cls.excel_df = dataframes.df_smiles_2_mols(cls.excel_df, 'Smiles', 'Molecule')
        cls.sdf_df = PandasTools.LoadSDF('test/assets/sdf_test_case.sdf', molColName='Molecule')
        cls.sdf_df.ID = list(range(4))
        
        return super().setUpClass()
    
    def test_df_smiles_2_mols(self):
         # Test with dropna
        out1 = dataframes.df_smiles_2_mols(self.mol_test_df, 'SMILES', 'ROMol')
        assert out1.ROMol.map(Chem.MolToSmiles, na_action='ignore').equals(
            self.mol_test_df.Molecule.map(Chem.MolToSmiles, na_action='ignore').dropna())
        
        # Test without dropna
        out2 = dataframes.df_smiles_2_mols(self.mol_test_df, 'SMILES', 'ROMol', dropna=False)
        assert out2.ROMol.map(Chem.MolToSmiles, na_action='ignore').equals(
            self.mol_test_df.Molecule.map(Chem.MolToSmiles, na_action='ignore'))
    
    def test_df_mols_2_inchi_keys(self):
        # Test with dropna
        out1 = dataframes.df_mols_2_inchi_keys(self.mol_test_df, 'Molecule', 'inchi')
        assert out1.inchi.equals(self.mol_test_df.InChi.dropna())
        
        # Test without dropna
        out2 = dataframes.df_mols_2_inchi_keys(self.mol_test_df, 'Molecule', 'inchi', dropna=False)
        assert out2.inchi.equals(self.mol_test_df.InChi)
        
    def test_df_mols_2_smiles(self):
        # Test with dropna
        out1 = dataframes.df_mols_2_smiles(self.mol_test_df, 'Molecule', 'smiles')
        assert out1.smiles.equals(self.mol_test_df.SMILES.dropna())
        
        # Test without dropna
        out2 = dataframes.df_mols_2_smiles(self.mol_test_df, 'Molecule', 'smiles', dropna=False)
        assert out2.smiles.equals(self.mol_test_df.SMILES)
        
    def test_write_sdf(self):
        # Write new and load new
        dataframes.write_sdf(self.excel_df, 'test/assets/sdf_test_out.sdf', mol_col_name='Molecule')
        load_sdf = PandasTools.LoadSDF('test/assets/sdf_test_out.sdf', molColName='Molecule')
        
        # Ensure Mols are the same
        self.assertEqual(
            naclo.mols_2_smiles(load_sdf.Molecule),
            naclo.mols_2_smiles(self.sdf_df.Molecule)
        )
        
        # Ensure original SMILES are the same
        assert load_sdf.Smiles.equals(self.sdf_df.Smiles)


class TestDataframesWriter(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_df = pd.DataFrame({
            'SMILES': ['O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1', 
                       'O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1', 
                       'CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O'],
            'Molecule': [Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1'),
                         Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1'),
                         Chem.MolFromSmiles('CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O')]
            })
        
        cls.writer = dataframes.Writer(df=cls.test_df, mol_col_name='Molecule')
        return super().setUpClass()
    
    def test_writer_rdkit_2_excel(self):
        # Test w/ buffer
        # buf = BytesIO()
        # self.writer.rdkit_2_excel(buf)
        
        # Test with path --> REQUIRES VISUAL CONFIRMATION
        path = 'test/assets/writer_rdkit_2_excel_test_case.xlsx'
        self.writer.rdkit_2_excel(path)
        
    def test_writer_write(self):
        buf = BytesIO()
        for ext in ['csv', 'tsv', 'xls', 'xlsx', 'sdf']:
            out = self.writer.write(buf, ext=ext)
            assert isinstance(out, BytesIO)


if __name__ == '__main__':
    unittest.main()
